from os import environ

import tomllib
from pydantic import BaseModel

BASE_URL = 'https://api.cimediacloud.com/'
TOKEN_URL = 'https://api.cimediacloud.com/oauth2/token'


class Config(BaseModel):
    ENV_PREFIX: str = 'CI_'
    TOML_KEY: str = 'sonyci'

    base_url: str = BASE_URL
    token_url: str = TOKEN_URL
    username: str | None = None
    password: str | None = None
    client_id: str | None = None
    client_secret: str | None = None
    workspace_id: str | None = None

    @classmethod
    def load(
        cls,
        toml_filename: str | None = None,
        toml_key: str = TOML_KEY,
        env_prefix: str = ENV_PREFIX,
        **kwargs,
    ):
        """
        Returns a new Config instance loaded with data in the following precedence: arguments,
        environment variables, toml file, defaults defined in the Config class.
        """
        if toml_filename:
            from_toml = cls.from_toml(filename=toml_filename, key=toml_key)
        else:
            from_toml = {}
        from_env = cls.from_env(env_prefix)
        params = {**from_toml, **from_env, **kwargs}
        return cls(**params)

    @classmethod
    def from_toml(cls, filename: str, key: str = TOML_KEY) -> dict:
        with open(str(filename), 'rb') as f:
            vals = tomllib.load(f)
        if key:
            return vals.get(key)
        return vals

    @classmethod
    def load_from_toml(cls, filename: str, key: str = TOML_KEY):
        return cls(**cls.from_toml(filename=filename, key=key))

    @classmethod
    def from_env(cls, prefix: str = ENV_PREFIX) -> dict:
        if not prefix:
            return {}

        # Returns a dict of config vars from the environment that begin with `prefix`
        return {
            k[len(prefix) :].lower(): v
            for k, v in environ.items()
            if k.startswith(prefix)
        }



from json import dumps, loads
from pathlib import Path
from typing import Optional
from urllib.request import urlretrieve

from requests_oauth2client.tokens import BearerToken, BearerTokenSerializer
from typer import Argument, Context, Exit, Option, Typer
from typer.main import get_group
from typing_extensions import Annotated

from sonyci import SonyCi
from sonyci.log import log
from sonyci.types import ProxyType
from sonyci.utils import save_token_to_file

app = Typer(context_settings={'help_option_names': ['-h', '--help']})


def parse_bearer_token(token: str) -> BearerToken:
    """Parse a bearer token from a json string."""
    return BearerToken(loads(token))


def version_callback(value: bool):
    """Print the version of the program and exit."""
    if value:
        from ._version import __version__

        print(f'v{__version__}')

        raise Exit()


class ProxyNotFoundError(Exception):
    """Raised when a specific proxy is not found."""


try:
    from trogon import Trogon

    @app.command()
    def tui(ctx: Context):
        Trogon(get_group(app), click_context=ctx).run()

except ImportError:
    pass


@app.command()
def login(
    ctx: Context,
    username: str = Option(
        ..., '--username', '-u', help='Sony CI username.', envvar='CI_USERNAME'
    ),
    password: str = Option(
        ..., '--password', '-p', help='Sony CI password.', envvar='CI_PASSWORD'
    ),
    test: bool = Option(False, '--test', '-t', help='Skips saving the token.'),
):
    """Login to Sony CI."""
    from sonyci.utils import get_token

    token: BearerToken = get_token(
        username,
        password,
        ctx.parent.params.get('client_id'),
        ctx.parent.params.get('client_secret'),
    )
    if not test:
        save_token_to_file(token, '.token')
    log.success('logged in to Sony CI!')


@app.command()
def get(ctx: Context, path: Annotated[str, Argument(..., help='The path to GET')]):
    """Make a GET request to Sony CI."""
    ci = SonyCi(t=ctx.parent.params['token'])
    log.trace(f'GET {path}')
    result = ci(path)
    log.success(result)
    print(dumps(result))


@app.command()
def post(
    ctx: Context,
    path: Annotated[str, Argument(..., help='The path to POST')],
    data=Argument(help='The data to POST'),
):
    """Make a POST request to Sony CI."""
    ci = SonyCi(t=ctx.parent.params['token'])
    data = loads(data)
    log.debug(f'POST {path} {data}')
    result = ci.post(path, data)
    log.success(result)
    print(dumps(result))


@app.command()
def search(
    ctx: Context, query: Annotated[str, Argument(..., help='The query to search for')]
):
    """Search for files in a Sony CI workspace"""
    ci = SonyCi(
        t=ctx.parent.params['token'], workspace_id=ctx.parent.params['workspace_id']
    )
    log.trace(f'search {query}')
    result = ci.workspace_search(query)
    log.success(result)
    print(dumps(result))


@app.command()
def download(
    ctx: Context,
    id: Annotated[str, Argument(..., help='The SonyCi ID of the file to download')],
    proxy: Annotated[ProxyType, Option('--proxy', '-p', help='Download ')] = None,
    output: Annotated[
        Optional[Path],
        Option('--output', '-o', help='The path to download the file to'),
    ] = None,
):
    """Download a file from Sony CI"""
    ci = SonyCi(t=ctx.parent.params['token'])
    log.trace(f'download id: {id} proxy: {proxy} output: {output}')
    result = ci.asset_download(id)
    link = result['location']
    if proxy:
        for p in result['proxies']:
            if p['type'] == proxy.value:
                log.debug(f'found proxy {proxy.value}')
                link = p['location']
                break
        # Check if matching proxy was found. Raise an exception if not found.
        if link == result['location']:
            raise ProxyNotFoundError(f'proxy {proxy} not found')

    log.trace(f'link: {link}')
    filename = output or Path(link).name.split('?')[0]
    log.debug(f'downloading {id} to {filename}')
    urlretrieve(link, filename)
    log.success(f'downloaded {id} to {filename}')


@app.command()
def asset(
    ctx: Context,
    asset: Annotated[str, Argument(..., help='The asset ID to search for')],
):
    """Search for files in a Sony CI workspace"""
    ci = SonyCi(
        t=ctx.parent.params['token'], workspace_id=ctx.parent.params['workspace_id']
    )
    log.trace(f'asset {asset}')
    result = ci.asset(asset)
    log.success(result)
    print(dumps(result))


@app.callback()
def main(
    ctx: Context,
    version: bool = Option(
        None,
        '--version',
        '-V',
        callback=version_callback,
        is_eager=True,
        help='Show the version and exit.',
    ),
    verbose: bool = Option(None, '--verbose', '-v', help='Show verbose output.'),
    token: Annotated[
        BearerToken,
        Option(
            '--token',
            '-t',
            parser=parse_bearer_token,
            help='Sony CI token.',
            envvar='CI_TOKEN',
        ),
    ] = None,
    workspace_id: str = Option(
        None,
        '--workspace-id',
        '-w',
        help='Sony CI workspace ID.',
        envvar='CI_WORKSPACE_ID',
    ),
    client_id: str = Option(
        None, '--client-id', '-c', help='Sony CI client ID.', envvar='CI_CLIENT_ID'
    ),
    client_secret: str = Option(
        None,
        '--client-secret',
        '-s',
        help='Sony CI client secret.',
        envvar='CI_CLIENT_SECRET',
    ),
):
    if not verbose:
        log.remove()
    if not token:  # and if command is not login
        log.debug('no token provided, trying to load from .token')
        try:
            with open('.token') as f:
                token = BearerTokenSerializer().loads(f.read())
                log.debug('loaded token from .token')
                ctx.params['token'] = token
        except FileNotFoundError:
            log.debug('no .token file found')


if __name__ == '__main__':
    app()



from loguru import logger as log

try:
    from rich.logging import RichHandler

    log.configure(
        handlers=[
            {
                'sink': RichHandler(),
            }
        ]
    )

except ImportError:
    pass



from os import environ

from pytest import fixture

from sonyci import Config


@fixture(
    scope='module',
    params=[
        {'filename': './tests/sonyci/sonyci.toml'},
        {'filename': './tests/sonyci/sonyci_no_key.toml', 'key': None},
        {
            'filename': './tests/sonyci/sonyci_different_key.toml',
            'key': 'different_key',
        },
    ],
)
def dict_from_toml(request):
    """Parameterized fixture returning dict of values from given toml file, under the given toml key"""
    return Config.from_toml(**request.param)


def test_from_toml(dict_from_toml):
    """Returns a dict of config values from a given toml file under a given toml key"""
    assert dict_from_toml['username'] == 'test username'
    assert dict_from_toml['password'] == 'test password'
    assert dict_from_toml['client_id'] == 'test client id'
    assert dict_from_toml['client_secret'] == 'test client secret'


def test_load_precedence():
    """Returns a Config instance loaded with data merged from different sources"""
    environ['CI_USERNAME'] = 'username from environment'
    config = Config.load(
        toml_filename='./tests/sonyci/sonyci.toml', password='password from arg'
    )
    assert config.username == 'username from environment'
    assert config.password == 'password from arg'
    assert config.client_id == 'test client id'



from json import loads

from pytest import mark

from sonyci.cli import app


@mark.vcr()
def test_asset_search(runner, asset_id):
    result = runner.invoke(app, ['asset', asset_id])
    assert result.exit_code == 0
    asset = loads(result.output)
    assert isinstance(asset, dict)
    assert asset['id'] == asset_id



from requests import post
from requests_oauth2client.tokens import BearerToken, BearerTokenSerializer

from sonyci.config import TOKEN_URL
from sonyci.log import log


def get_token(
    username: str,
    password: str,
    client_id: str,
    client_secret: str,
    token_url: str = TOKEN_URL,
) -> BearerToken:
    """Login to SonyCI and return a BearerToken."""
    response = post(
        token_url,
        auth=(username, password),
        data={
            'grant_type': 'password',
            'client_id': client_id,
            'client_secret': client_secret,
        },
    )
    log.debug(f'token response status: {response.status_code}')
    if response.status_code != 200:
        log.error(f'Token returned {response.status_code}: {response.text}')
        raise Exception(
            f'Token did not return 200. Returned: {response.status_code}: {response.text}'
        )
    return BearerToken(**response.json())


def get_token_from_file(filename: str = '.token'):
    with open(filename) as f:
        return BearerTokenSerializer().loads(f.read())


def save_token_to_file(token: BearerToken, filename: str = '.token'):
    with open(filename, 'w') as f:
        f.write(BearerTokenSerializer().dumps(token))


def json(func):
    """Decorator for calling .json() on Response objects."""

    def inner(*args, **kwargs):
        return func(*args, **kwargs).json()

    return inner
